# -*- coding: utf-8 -*-
"""
Created on Sat Apr 26 17:04:01 2014 - zeroEpochTime = datetime(1970,1,1)
@author: David Worrall
Version: 0.03 August 2014 for live streaming

SFLOW SONIFICATION 
Class 1: SflowSonifyer

sflow code parser code - Controls the (global) sonification process.
- read sflow data from an ascii pipe 'sink' (file)
- create a rec array for each line of sflow data
- open the data file
- read line at a time into a numpy recarray

It takes as input a real-time stream of sflow packet data and payloads, either as

a) a live stream from a named pipe ( self.sflowFIFO, hardwired as the FIFO 
'allsflowdata') which is piped to by the sflow packet capture software 
(self.FIFOfeedSW = RunCapture.x.x.x) which itself filters the sdata off the 
sflow switches., or
b) As a non-real-time file read of backup files from a)

"""

import csnd6

import numpy as np
#from datetime import datetime
import time
import glob
import os
import subprocess
import socket
import urllib 

from random import random
from random import choice

def cleanUserInput(message):
    if message[0] == '"':
        return message[1:-1].strip() # strip any quotes from input string
    else:
        return message
        
def dictFromString(stringy):
    # makes a (pair, value) dict from a string sequence.
    # eg if s="1\tgreeting\n5\tgrub\n7\techo\n"
    # dictFromString (s) returns {'1': 'greeting', '5': 'grub', '7': 'echo'}
    stringy=stringy.replace("\t", ' ').replace('\n', ' ').split()
    return  dict( zip(stringy[0:len(stringy):2],stringy[1:len(stringy):2]))
    

class SflowSonifyer(object):
    def __init__(self, dataDirectory): # "/Users/wor-local/sonipy/_CorpoReal_data/internet/sflowCapture/")
        """self.filenameStem  = '/Users/wor/sonipy/edf/wfdp/ecgtk-master/samples/format212/100'
        # the directory where the sink files, or named pipes resides and from ehere it is sonified."""
#        self.backupfile = filename
        self.dataDirectory = dataDirectory # "/Users/wor-local/sonipy/_CorpoReal_data/internet/sflowCapture/"
        self.sflowBackupDir = self.dataDirectory+"backup/"
        self.testFile=""
#        self.testFile="/Users/wor-local/sonipy/_CorpoReal_data/internet/sflowCapture/backup/99_collectedData_09.csv"
        self.sinkFileNames = glob.glob(self.sflowBackupDir+"*.csv")
        self.FIFOfeedSW = self.dataDirectory+"RunCapture_2014-05-13_03.jar" # the sflow Packet capture SW
        self.fileID = None # file descriptor. Used for sonifying backup data

        self.FIFO=self.dataDirectory+"allsflowdata"
        self.fifoID = None # pipe descriptor. Used for sonifying live FIFO data

        self.activeIPaddrs = {} # a dict of active IP addresses derived from the .jar config file
        self.JarConfigfile =self.dataDirectory+"config.properties"
#        self.JarConfigfile ="config.properties" # DANGEROUS! should use the above - trransfer when done!
        self.IPdict = {} # a dict in which keys are tupes of ip address. Initialised in self.loadActiveIPaddrs
        self.unidentifiedIPaddr =[] # used mainly for testing whether an address is actually external
        self.suspectIPaddr =[] # used mainly for testing whether an address is actually external
        self.printerRecords =[] # checking whether we ar picking them up
        self.printRate = 60 # rate at which Records,when print == True, will be printed.

        self.portsIDfile="./TCPandUDPportIDs.txt"  # In CWD by default
        self.portsDict ={}
        self.unidentifiedPorts = []

        self.sflowDataFieldSeparator=";"
        self.packetFields="sflowPacketNumber;samplePacketNumber;receivingTimestamp;IPprotocol"
        self.packetFields+=";macSrc;macDst;ipAdrSrc;ipAdrDst;tcpIPportSrc;tcpIPportDst;tcpSeq"
        self.packetFields+=";tcpAck;udpIPportSrc;udpIPportDst;extractionTimestamp;sflowSeqNumber"
        self.packetFields+=";sflowSysUptime;sflowSamplingRate;sflowSequenceNumber"
        self.packetFields=self.packetFields.split(self.sflowDataFieldSeparator) # packet fields as a list
        self.lenPacketFieldsList=len(self.packetFields)
        self.packetFieldsDict = dict(zip (self.packetFields, range(self.lenPacketFieldsList))) # a {name:index} dict
        self.identityMaskIndexList=[] # A list of Indexes ('fields') of the packet list for testing for things like redundancy
        self.previousMaskedRecord =[] # a memory buffer of THE RESULTS of applying the self.identityMaskIndexList
                # to the previous record. # Used to check if consecutive records are substantially different
        self.recordList = []  # the big daddy list that contains the current read before it is sound rendered
        self.playList = [] # the list of events lists to render to sound
        self.live = False
        self.printORnot = False # used to switch the array print routines on or off
        self.sendUDP_IP = ''
        self.sendUDP_PORT = 0
        self.sock = ''
        self.UPDwelcomeMessage = "drw: G'day Mate!"

        self.sflowStatsDict = {} #  dictionary of things to count in the passing sflow stream. See initStatsDict()

    def createPacketIdentityMasks(self):
        # create a list of indexes ('fields') of the packet list for testing for things like redundancy
        self.identityMaskIndexList=[
            # self.packetFieldsDict['receivingTimestamp']
            # self.packetFieldsDict['sflowPacketNumber']
            #  self.packetFieldsDict['ipAdrSrc'], # record indexes
            #  self.packetFieldsDict['ipAdrDst'], # to check for identity
            #  self.packetFieldsDict['tcpIPportSrc'], # btwn consec. records
            #  self.packetFieldsDict['tcpIPportDst']
        ] 

    def loadActiveIPaddrs(self):
        #Parses the JarConfigfile containing the IP addresses to be parsed for.
        # Warning: need to check that the actual code is active
        descriptors = []
        kees = [] #  
        for line in open(self.JarConfigfile):
            if "#.x" in line:    # takes only the descriptor lines, not the active code
                splitline = line.rstrip('\n').replace('-', '').replace('.x','').split()[1:] # [:-1]cleanup fiddles
                kees += [tuple (splitline[:-1])]
                descriptors += [splitline [-1]]
        self.IPdict = dict(zip (kees, descriptors))       
        
    def loadPortIDs(self):
        #loads the IP port ids from a file. Hardwired as self.portsIDfile=TCPandUDPportIDs.txt in the current directory.
        try:
            fid=open(self.portsIDfile, 'r') # Open it again for readline?
            temp=fid.read()
            self.portsDict = dictFromString(temp)
        except:
            print "Problem with reading the IP ports ID file."
        finally:
            fid.close()

    def printSflowRecord(self,recordNr, recordString):
        # prints the record as as list.
        print "Record Nr: ", recordNr, "\n=================="
        for fieldNr in range(self.lenPacketFieldsList):
            print "%20s %s" % (self.packetFields[fieldNr], recordString[fieldNr] )

    def IPaddrID (self, stringy):
        #resolves an IP string to a location according to the current config file. Unfinished... 
        for kee in self.IPdict.keys():  # kee is a tuple of possible ip addresses
            if stringy.startswith(kee):
                return self.IPdict[kee] # the description
            if stringy.startswith(('10.54', '10.55','153.96')):    # temp code to resolve unknowns
                if stringy not in self.suspectIPaddr:
                    self.suspectIPaddr +=  [stringy]
            else:
                if stringy not in self.unidentifiedIPaddr:
                    self.unidentifiedIPaddr += [stringy]
#        return stringy     # couldn't find the IPaddress string
        return "_unidentifiedIP"     # couldn't find the IPaddress string

    def IPportID (self, stringy):
        #resolves a port string to an ID according to self.PortsDict which uses the self.portsIDfile file
        for kee in self.portsDict.keys():  # kee is a tuple of possible ip addresses
            if stringy==kee:
                return self.portsDict[kee] # the description - DOESN'T SEEM to RETURN
                break
            if stringy not in self.unidentifiedPorts:  # if it can't find the descriptor
                self.unidentifiedPorts += [stringy]
#        return stringy     # couldn't find the IPaddress string
        return "_unidentifiedPortID"     # couldn't find the IPaddress string

    def printRecordArray(self, recordNr, recordList):
        # prints the info needed to make early decisions - in development
        # Only active when self.printORnot = True
        h1="  rec  ptcl  recTime        sfPNr    spPNr      SrcIP               DestIP "
        h2="           tcpSPrt     tcpDPrt      udpSPrt udpDPrt\n"
        if (recordNr % 40)==0:       
            print h1+h2 + "="*len (h1+h2)
#        for a single record as a list of string items:
        print "%5s %3s%15s%10s%10s%20s%19s%12s%12s%12s%12s" % ( recordNr,
                          recordList[self.packetFieldsDict['IPprotocol']],
                          recordList[self.packetFieldsDict['receivingTimestamp']],
                          recordList[self.packetFieldsDict['sflowPacketNumber']],
                          recordList[self.packetFieldsDict['samplePacketNumber']],
                    self.IPaddrID(recordList[self.packetFieldsDict['ipAdrSrc']]),
                    self.IPaddrID(recordList[self.packetFieldsDict['ipAdrDst']]),
                    self.IPportID(recordList[self.packetFieldsDict['tcpIPportSrc']]),
                    self.IPportID(recordList[self.packetFieldsDict['tcpIPportDst']]),
                    self.IPportID(recordList[self.packetFieldsDict['udpIPportSrc']]),
                    self.IPportID(recordList[self.packetFieldsDict['udpIPportDst']])
                          )
        time.sleep(60.0/self.printRate) #60 by default. i.e. 1/sec. 

    def extractSingleRecordFromFile(self, recordNr):
        # extracts a single record string from the current file. RecordNr is file lineNr.
        # NB it is passed to self.printSflowRecord as a string.
        fid=open(self.testFile, 'r') # Open it again for readline?
        try:
            for i in range (recordNr):
                fid.readline()      # burn off all lines preceding the one we want
            line=fid.readline()     # now read the line we want to display
            self.printSflowRecord(recordNr, record.replace(';;',';-;').split(
                        self.sflowDataFieldSeparator))
        except:
            print "Sorry, there is not that many records in the file."
        finally:
            fid.close()

    def isRecordRedundant(self, record):
#        return False
        # Determines whether the record is redundant. i.e. not for sound rendering
        self.createPacketIdentityMasks()  # applies various identity masks in the self.identityMaskIndexList
        maskValues = [record[index] for index in self.identityMaskIndexList]
        if  maskValues == self.previousMaskedRecord:
            self.previousMaskedRecord = maskValues  # is this necessary, given they are equivalent?
            return True
        else:
            self.previousMaskedRecord = maskValues
            return False

# def readSflowFile goes here
    def readSflowFile2RecordList(self):
        if not self.fileID:
            self.fileID=open(self.testFile)
            print "opening %s for reading" % self.testFile
        linecount=0
        tc1=time.clock()
        t1=time.time()

        self.fileID.seek(0) #  rewind the Fptr in preparation for reading the whole file
        try:
            self.recordList=self.fileID.readlines()  #read the available records into the array
            for indx in range (len( self.recordList)):
                self.recordList[indx] = self.recordList[indx].replace(';;',';-;').split(self.sflowDataFieldSeparator)
                tmp = (int(self.recordList[indx][self.packetFieldsDict['receivingTimestamp']])/1000.0) # *10 # convert sec (and truncate to nearest 0.1sec)
#                tmp = int(self.recordList[indx][self.packetFieldsDict['receivingTimestamp']])/10000.0) *10 # convert ms to sec and truncate to nearest sec
                self.recordList[indx][self.packetFieldsDict['receivingTimestamp']]= tmp
#                self.recordList[indx][self.packetFieldsDict['receivingTimestamp']]= int(
#                            self.recordList[indx][self.packetFieldsDict['receivingTimestamp']])/1000 # convert ms to sec and truncate to nearest sec
                if self.isRecordRedundant(self.recordList[indx]) == True:
                    self.recordList[indx][self.packetFieldsDict['receivingTimestamp']]= int(self.recordList[indx][self.packetFieldsDict['receivingTimestamp']])
                    self.recordList[indx][-1]='*'  # replace the last field with a '*' so we can see it is redundant on printing
                if self.printORnot: # print  self.recordList[self.packetFieldsDict['IPprotocol']] # = "*" self.recordList was "record"
                    self.printRecordArray(linecount, self.recordList[indx])  # self.recordList was "record"
                linecount += 1
                #::: PAUSE SWITCH ::::::::: PAUSE SWITCH ::::::::#
                #if pause !=0:
                #    time.sleep(pause)
        finally:
            self.fileID.close()
        t2 = time.time()
        tc2=time.clock()
        print t2-t1, tc2-tc1

    def makePlayList(self):
        # makes a play list from the record List ( a list of strings) by converting it to a list of lists
        #startTimeOffset=int(self.recordList[0].split(self.sflowDataFieldSeparator)[2])
        startTimeIndex=self.packetFieldsDict['receivingTimestamp']
        for record in self.recordList:
            self.playList +=[record]  #a list of lists
            
# :::::::::::::::::::PACKET ANALYSIS FUNCTIONS :::::::::::
    def initStatsDict (self):
        # initialised the self.sflowStatsDict for things to count in the incoming sflow stream
        # unfinished - portsDict inti works but IPdict doesn't,  because the keys are tuples of IPaddresses
        try:
            for key in self.IPdict.keys():
                print key
                self.sflowStatsDict.update({'src'+key:0})
                self.sflowStatsDict.update({'dst'+key:0})
        except:
            print "Unable to initialise the Stats dictionary with IP addresses"
        try:
            for key in self.portsDict.keys():
                self.sflowStatsDict.update({'src'+key:0})
                self.sflowStatsDict.update({'dst'+key:0})
            self.sflowStatsDict.update({'unidentifiedIPaddr':[]}) # a list of unidentified IP addresses - external comms?
        except:
            print "Unable to initialise the Stats dictionary with Ports info."

    def updateStatsDict (self, record):
        # updates self.sflowStatsDict elements. If an element is not in the dict, it adds it automatically
        srcIP ='ipAdrSrc'+self.IPaddrID(record[self.packetFieldsDict['ipAdrSrc']])
        if srcIP in self.sflowStatsDict.keys():
            self.sflowStatsDict.update({srcIP:self.sflowStatsDict[srcIP]+1})
        else:
            self.sflowStatsDict.update({srcIP:1})
            
        dstIP ='ipAdrDst'+self.IPaddrID(record[self.packetFieldsDict['ipAdrDst']])
        if dstIP in self.sflowStatsDict.keys():
            self.sflowStatsDict.update({dstIP:self.sflowStatsDict[dstIP]+1})
        else:
            self.sflowStatsDict.update({dstIP:1})

        # ----------  the TCP ports:
        # firstly, record the TCP Src port number:
        srcPortNr = record[self.packetFieldsDict['tcpIPportSrc']]   # get the raw port number
        portNrIDStr = 'tcpPortSrc_'+str(srcPortNr)                # a string name for the port - to use as a dict key NB the name change 'tcpPortSrc_'
        if portNrIDStr in self.sflowStatsDict.keys():               # save the raw port number as a dict key
            self.sflowStatsDict.update({portNrIDStr:self.sflowStatsDict[portNrIDStr]+1}) # increment the durrent value by 1
        else:                                                       # not in the dict, so put one there
            self.sflowStatsDict.update({portNrIDStr:1})
        # secondly, record the TCP Src port ID as per the:
        srcPort ='tcpPortSrc'+self.IPportID(srcPortNr)    # IP port ID string
#        srcPort ='tcpIPportSrc'+self.IPportID(record[self.packetFieldsDict['tcpIPportSrc']])        
        if srcPort in self.sflowStatsDict.keys():
            self.sflowStatsDict.update({srcPort:self.sflowStatsDict[srcPort]+1})
        else:
            self.sflowStatsDict.update({srcPort:1})
        # end of TCP Src port recording
        # now, record the TCP Dst port number:            
        dstPortNr = record[self.packetFieldsDict['tcpIPportDst']]   # get the raw port number
        portNrIDStr = 'tcpPortDst_'+str(dstPortNr)                # a string name for the port - to use as a dict key NB the name change 'tcpPortDst_'
        if portNrIDStr in self.sflowStatsDict.keys():               # save the raw port number as a dict key
            self.sflowStatsDict.update({portNrIDStr:self.sflowStatsDict[portNrIDStr]+1}) # increment the durrent value by 1
        else:                                                       # not in the dict, so put one there
            self.sflowStatsDict.update({portNrIDStr:1})
        # secondly, record the TCP Src port ID as per the control file:
        dstPort ='tcpIPportDst'+self.IPportID(dstPortNr)    # IP port ID string
#        dstPort ='tcpIPportDst'+self.IPportID(record[self.packetFieldsDict['tcpIPportDst']])        
        if dstPort in self.sflowStatsDict.keys():
            self.sflowStatsDict.update({dstPort:self.sflowStatsDict[dstPort]+1})
        else:
#            self.sflowStatsDict.update({srcPort:1})  #   drw: BUG?  20150122
            self.sflowStatsDict.update({dstPort:1})

        # end of TCP Src port recording
#TCP STUFF here
# --- new
        # ------------- the UDP ports:
        # firstly, record the UDP Src port number:
        PortNr = record[self.packetFieldsDict['udpIPportSrc']]   # get the raw port number
        portNrIDStr = 'udpPortSrc_'+str(srcPortNr)                # a string name for the port - to use as a dict key
        if portNrIDStr in self.sflowStatsDict.keys():               # save the raw port number as a dict key
            self.sflowStatsDict.update({portNrIDStr:self.sflowStatsDict[portNrIDStr]+1}) # increment the durrent value by 1
        else:                                                       # not in the dict, so put one there
            self.sflowStatsDict.update({portNrIDStr:1})
        # secondly, record the UDP Src port ID as per the control file:
        srcPort ='udpIPportSrc'+self.IPportID(srcPortNr)    # IP port ID string
#        srcPort ='udpIPportSrc'+self.IPportID(record[self.packetFieldsDict['udpIPportSrc']])        
        if srcPort in self.sflowStatsDict.keys():
            self.sflowStatsDict.update({srcPort:self.sflowStatsDict[srcPort]+1})
        else:
            self.sflowStatsDict.update({srcPort:1})

        # end of UDP Src port recording
        # now, record the UDP Dst port number:            
        dstPortNr = record[self.packetFieldsDict['udpIPportDst']]   # get the raw port number
        portNrIDStr = 'udpPortDst_'+str(dstPortNr)                # a string name for the port - to use as a dict key
        if portNrIDStr in self.sflowStatsDict.keys():               # save the raw port number as a dict key
            self.sflowStatsDict.update({portNrIDStr:self.sflowStatsDict[portNrIDStr]+1}) # increment the durrent value by 1
        else:                                                       # not in the dict, so put one there
            self.sflowStatsDict.update({portNrIDStr:1})
        # secondly, record the TCP Src port ID as per the:
        dstPort ='udpIPportDst'+self.IPportID(dstPortNr)    # IP port ID string
#        dstPort ='udpIPportDst'+self.IPportID(record[self.packetFieldsDict['udpIPportDst']])        
        if dstPort in self.sflowStatsDict.keys():
            self.sflowStatsDict.update({dstPort:self.sflowStatsDict[dstPort]+1})
        else:
#            self.sflowStatsDict.update({srcPort:1})  #   drw: BUG?  20150122
            self.sflowStatsDict.update({dstPort:1})
        # end of UDP Src port recording

# -- end of new

"""
TCP STUFF

        dstPortNr = record[self.packetFieldsDict['tcpIPportDst']]  # the raw port numer
        dstPort ='tcpIPportDst'+self.IPportID(dstPortNr)
#        dstPort ='tcpIPportDst'+self.IPportID(record[self.packetFieldsDict['tcpIPportDst']])
        if dstPortNr in self.sflowStatsDict.keys():
            self.sflowStatsDict.update({dstPortNr:self.sflowStatsDict[srcPortNr]+1})

        if dstPort in self.sflowStatsDict.keys():
            self.sflowStatsDict.update({dstPort:self.sflowStatsDict[dstPort]+1})
        else:
            self.sflowStatsDict.update({dstPort:1})



        # the UDP ports:
        srcPort ='udpIPportSrc'+self.IPportID(record[self.packetFieldsDict['udpIPportSrc']])
        if srcPort in self.sflowStatsDict.keys():
            self.sflowStatsDict.update({srcPort:self.sflowStatsDict[srcPort]+1})
        else:
            self.sflowStatsDict.update({srcPort:1})

        dstPort = 'udpIPportDst'+self.IPportID(record[self.packetFieldsDict['udpIPportDst']])
        if dstPort in self.sflowStatsDict.keys():
            self.sflowStatsDict.update({dstPort:self.sflowStatsDict[dstPort]+1})
        else:
               self.sflowStatsDict.update({dstPort:1})

        #print "HERE WE ARE\n"
"""        
